use IO::Socket::Async::SSL;
use URI;

use Ramini::Logger;

unit class Ramini::Server;

has Int $.port;
has Str $.listen;

has Str $.root;
has Str $.domain;

has Str $.certificate-file is required;
has Str $.private-key-file is required;

method run() {
    my $connections = IO::Socket::Async::SSL.listen($!listen, $!port, :$.certificate-file, :$.private-key-file,
                                                    :no-session-resumption-on-renegotiation(False));
    react {
        my $listener = do whenever $connections -> $socket {
            start self!handle-connection($socket)
        }

        say "Server listening on port " ~ $!port;

        whenever signal(SIGINT) {
            say "Shutting down...";
            $listener.close;
            exit;
        }
    }
}

method !write-response($conn, $status, $meta, $log, :$body-stream) {
    my $header = $status.Str ~ " " ~ $meta;
    $log.log($header);
    await $conn.print($header ~ "\r\n");
    $body-stream.tap: { await $conn.write($_) } if $body-stream;
}

method !serve-file($conn, $path, $log) {
    $log.log("Serving " ~ $path);
    self!write-response($conn, 20, "", $log,
                        body-stream => $path.open(:bin).Supply)
}

method !handle-request($conn, $request, $log) {
    my $file = $.root.IO.add($request.path);
    when $request.host !~~ $.domain {
        self!write-response($conn, 53, "This server does not accept proxy requests", $log)
    }
    when $file.IO ~~ :f { self!serve-file($conn, $file.IO, $log) }
    when $file.IO ~~ :d && $file.add("index.gmi").IO ~~ :f {
        self!serve-file($conn, $file.add("index.gmi").IO, $log);
    }
    default { self!write-response($conn, 51, "File not found", $log) }
}

method !handle-connection($conn) {
    my $start = now;
    my $buf = Buf.new;
    my $log = Ramini::Logger.new();
    $log.log("Incoming connection");
    react {
        whenever $conn.Supply(:bin) {
            $buf.append($_);

            # URL can be up to 1024 bytes long + CRLF
            my $req = try $buf.subbuf(0..1025).decode('utf-8');

            when $req && $req.contains("\r\n") {
                my $uri = URI.new($req.lines[0].chomp);
                self!handle-request($conn, $uri, $log);
                $conn.close;
            }

            when $buf.bytes > 1024 {
                self!write-response($conn, 59, "Request URL too long", $log);
                $conn.close;
            }
        }
    }
    CATCH {
        default {
            $log.exception($_);
            $conn.close;
        }
    }
    $log.log("Closed connection after " ~ (now - $start) ~ " seconds");
}
